--
-- Body
--
Body = {}
BodyMT = { __index = Body }

function Body:Create ( world, position, angle, linearDamping, angularDamping, fixedRotation, isBullet, allowSleep, isSleeping )
	local body = {}
	setmetatable ( body, BodyMT )

	body.world = world

	local def = b2BodyDef ( )
	def.position = position
	def.angle = angle
	def.linearDamping = linearDamping
	def.angularDamping = angularDamping
	def.fixedRotation = fixedRotation
	def.isBullet = isBullet
	def.allowSleep = allowSleep
	def.isSleeping = isSleeping

	body.def = def
	body.scale = 1
	body.shapes = {}

	body.sprite = Sprite ( position.x, position.y )
	body.sprite:set_rotation_r ( -angle )

	if body:Generate ( ) == false then
		return nil
	end

	body:Paint ( )

	return body
end

function Body:Destroy ( )
	self:DestroyAllShapes ( )
	self:DestroyAllJoints ( )

	self:Degenerate ( )

    if self.image then
        self.image = nil
        self.imageX = nil
        self.imageY = nil
    end
	self.sprite = nil

	self.def = nil
	self.scale = nil
	self.shapes = nil

	self.world = nil
end

function Body:Generate ( )
	self:Degenerate ( )

	self.contacts = {}
	self.collisions = {}

	self.b2body = self.world.b2world:CreateBody ( self.def )
	if self.b2body == nil then
		return false
	end
	self.b2body.parent = self

	self.world.layer:add_child ( self.sprite )

	local joints = self.world:GetAssociatedJoints ( self )
	for i, v in pairs ( joints ) do
		v:Generate ( )
	end
	for i, v in pairs ( self.shapes ) do
		v:Generate ( )
	end

	return true
end

function Body:Degenerate ( )
	if self.b2body == nil then
		return
	end

	local joints = self.world:GetAssociatedJoints ( self )
	for i, v in pairs ( joints ) do
		v:Degenerate ( )
	end
	for i, v in pairs ( self.shapes ) do
		v:Degenerate ( )
	end

	self.world.layer:remove_child ( self.sprite )

	self.b2body.parent = nil
	self.world.b2world:DestroyBody ( self.b2body )
	self.b2body = nil

	self.contacts = nil
	self.collisions = nil
end

function Body:DestroyAllShapes ( )
	while #self.shapes > 0 do
		local last = table.remove ( self.shapes )
		last:Destroy ( )
	end
	-- destroying all shapes does not zero out the body's mass
end

function Body:DestroyAllJoints ( )
	local joints = self.world:GetAssociatedJoints ( self )
	for i, v in pairs ( joints ) do
		self.world:DestroyJoint ( v )
	end
end

function Body:CreateShapeCircle ( radius, localPosition, density, friction, restitution, isSensor )
	local shape = ShapeCircle:Create ( self, radius, localPosition, density, friction, restitution, isSensor )

	if shape ~= nil then
		table.insert ( self.shapes, shape )
	end

	return shape
end

function Body:CreateShapePolygon ( vertices, density, friction, restitution, isSensor )
	local shape = ShapePolygon:Create ( self, vertices, density, friction, restitution, isSensor )

	if shape ~= nil then
		table.insert ( self.shapes, shape )
	end

	return shape
end

function Body:DestroyShape ( shape )
	shape:Destroy ( )

	for i, v in pairs ( self.shapes ) do
		if v == shape then
			table.remove ( self.shapes, i )
			break
		end
	end
end

function Body:GetScale ( )
	return self.scale
end

function Body:SetScale ( scale )
	if scale <= 0 then
		return
	end
	for i, v in pairs ( self.shapes ) do
		v:Degenerate ( )
		local type = v:GetType ( )
		if type == e_circleShape then
			v.def.radius = v.def.radius / self.scale * scale
			v.def.localPosition.x = v.def.localPosition.x / self.scale * scale
			v.def.localPosition.y = v.def.localPosition.y / self.scale * scale
		elseif type == e_polygonShape then
			for i, v in pairs ( v.vertices ) do
				v.x = v.x / self.scale * scale
				v.y = v.y / self.scale * scale
			end
		end
		v:Generate ( )
	end
	self.scale = scale
	self.sprite:set_scale ( scale, scale )
	--self:Paint ( )
end

-- stores the body's position and angle
function Body:Update ( )
    local b2body = self.b2body
	if b2body == nil or b2body:IsSleeping ( ) == true then
		return
	end
	
	local x = b2body:GetPositionX ( )
	local y = b2body:GetPositionY ( )
    self.def.position:Set ( x, y )
	self.def.angle = b2body:GetAngle ( )
end

--
-- Condition functions
--
-- checks if a given point lies inside any of the shapes of the body
function Body:TestPoint ( pt )
    if self.b2body then
        local xf = self.b2body:GetXForm ( )
        for i, v in pairs ( self.shapes ) do
            if v:TestPoint ( xf, pt ) == true then
                return true
            end
        end
    end
	return false
end

-- returns an adjusted linear velocity that ignores the effects of gravity and damping
function Body:GetCorrectedLV ( lv, seconds )
	-- deintegrate gravity
	local clv = b2Vec2 ( lv.x, lv.y )
	clv.x = clv.x - self.world.gravity.x * seconds
	clv.y = clv.y - self.world.gravity.y * seconds

	-- compensate damping
	local damping = math.clamp ( 1 - seconds * self.def.linearDamping, 0, 1 )
	if damping == 0 then
		clv:SetLength ( b2_maxLinearVelocity )
	else
		clv.x = clv.x / damping
		clv.y = clv.y / damping
	end

	return clv
end

-- returns an adjusted angular velocity that ignores the effects of damping
function Body:GetCorrectedAV ( av, seconds )
	-- compensate damping
	local damping = math.clamp ( 1 - seconds * self.def.angularDamping, 0, 1 )
	if damping == 0 then
		return b2_maxAngularVelocity
	else
		return av / damping
	end
end

-- returns the future position assuming no external forces are affecting the body
function Body:GetFuturePosition ( seconds )
	local lv = self.b2body:GetLinearVelocity ( )

	-- integrate gravity
	lv.x = lv.x + self.world.gravity.x * seconds
	lv.y = lv.y + self.world.gravity.y * seconds

	-- apply damping
	local damping = math.clamp ( 1 - seconds * self.def.linearDamping, 0, 1 )
	lv.x = lv.x * damping
	lv.y = lv.y * damping

	lv.x = lv.x * seconds + self.def.position.x
	lv.y = lv.y * seconds + self.def.position.y

	return lv
end

-- returns the future angle assuming no external forces are affecting the body
function Body:GetFutureAngle ( seconds )
	local av = self.b2body:GetAngularVelocity ( )

	-- apply damping
	local damping = math.clamp ( 1 - seconds * self.def.angularDamping, 0, 1 )
	av = av * damping

	av = av * seconds + self.def.angle

	return av
end

-- returns the time required to reach a given position
function Body:GetArriveTime ( position, speed )
	-- current position
	local currentPosition = self.def.position

	-- distance to target
    local distance = currentPosition:Distance ( position )

	-- time = distance / speed
	return distance / speed
end

-- returns the time required to rotate to a given angle
function Body:GetTurnTime ( angle, turnRate )
	-- current angle
	local a = self.def.angle

	-- angle to target
	local arc = angle - a

	-- time = arc / speed
	return arc / turnRate
end

-- returns the time required to reach a given linear velocity
function Body:GetAccelerationTimeLV ( fv, force )
	-- initial velocity
	local v = self.b2body:GetLinearVelocity ( )

	-- change in velocity
	v.x = fv.x - v.x
	v.y = fv.y - v.y

	-- mass
	local mass = self.b2body:GetMass ( )

	-- time = change in velocity * mass / force
	local time = v:Length ( ) * mass / force

	return time
end

-- returns the time required to reach a given angular velocity
function Body:GetAccelerationTimeAV ( fv, torque )
	-- initial velocity
	local iv = self.b2body:GetAngularVelocity ( )

	-- inertia
	local inertia = self.b2body:GetInertia ( )

	-- time = change in velocity * inertia / torque
	local time = ( fv - iv ) * inertia / torque

	return time
end

--
-- Locomotion
--
function Body:GetPosition ( )
	return self.def.position
end

function Body:SetPosition ( pt )
	self:SetPositionXY ( pt.x, pt.y )
end

function Body:SetPositionXY ( x, y )
	self.def.position:Set ( x, y )

	self.b2body:SetXForm ( self.def.position, self.def.angle )
end

function Body:GetAngle ( )
	return self.def.angle
end

function Body:SetAngle ( r )
	self.def.angle = r

	self.b2body:SetXForm ( self.def.position, r )
end

function Body:ApplyForce ( f, pt )
    self.b2body:ApplyForce ( f, pt or self.def.position )
end

function Body:ApplyImpulse ( i, pt )
    self.b2body:ApplyImpulse ( i, pt or self.def.position )
end

function Body:Move ( seconds, position, speed )
	local currentPosition = self.def.position

	-- condition
	local eta = self:GetArriveTime ( position, speed )
	if eta < seconds then
		self:SetPosition ( position )
		return eta
	end

	-- output
	-- vector to target
	local dist = currentPosition:Distance ( position )
	local step = speed * seconds
    local dx = ( position.x - currentPosition.x ) / dist * step
    local dy = ( position.y - currentPosition.y ) / dist * step
    self:SetPositionXY ( currentPosition.x + dx, currentPosition.y + dy )

	-- we are moving the body manually so clear all forces acting upon it
    --local lv = self:GetCorrectedLV ( b2_b2Vec2, seconds )
    --self.b2body:SetLinearVelocity ( lv )
    self.b2body:SetLinearVelocity ( b2_b2Vec2 )

	return seconds
end

function Body:MoveLV ( seconds, position, speed )
	local currentPosition = self.def.position

	-- condition
	local eta = self:GetArriveTime ( position, speed )
	if eta < seconds then
		self:SetPosition ( position )
		return eta
	end

	-- output
	-- vector to target
	local d = b2Vec2 ( position.x - currentPosition.x, position.y - currentPosition.y )
	d:SetLength ( speed )

	d = self:GetCorrectedLV ( d, seconds )
	self.b2body:SetLinearVelocity ( d )

	return seconds
end

function Body:MoveForce ( seconds, position, speed, maxForce, ease )
	local currentPosition = self.def.position
	local currentLV = self.b2body:GetLinearVelocity ( )
	local currentSpeed = currentLV:Length ( )

	-- condition
	if currentSpeed ~= 0 then
		local eta = self:GetArriveTime ( position, currentSpeed )
		if eta < seconds then
			self:SetPosition ( position )
			return eta
		end
	else
		if position.x == currentPosition.x and position.y == currentPosition.y then
			return 0
		end
	end

	-- output
	if ease == true then
		if currentSpeed ~= 0 then
			-- time required for a full stop
			local stopTime = self:GetAccelerationTimeLV ( b2_b2Vec2, maxForce )

			-- are we going to overshoot the target
			local eta = self:GetArriveTime ( position, currentSpeed )
			if eta <= stopTime then
				-- desired velocity = ( final position - initial position ) / time
				local d = self:GetFuturePosition ( stopTime )

				d.x = ( position.x - d.x ) / stopTime
				d.y = ( position.y - d.y ) / stopTime
				d = self:GetCorrectedLV ( d, seconds )

				-- force = ( velocity / time ) * mass
				local mass = self.b2body:GetMass ( )
				d.x = ( d.x / seconds ) * mass
				d.y = ( d.y / seconds ) * mass

				-- truncate to maximum steering force
				d:Truncate ( maxForce )
				self.b2body:ApplyForce ( d, currentPosition )

				return seconds
			end
		end
	end

	-- desired velocity
	local d = b2Vec2 ( position.x - currentPosition.x, position.y - currentPosition.y )
	d:SetLength ( speed )
	d = self:GetCorrectedLV ( d, seconds )

	self:AccelerateLV ( seconds, d, maxForce )

	return seconds
end

function Body:AccelerateLV ( seconds, desiredLV, maxForce )
	local currentPosition = self.def.position
	local currentLV = self.b2body:GetLinearVelocity ( )

	if currentLV.x == desiredLV.x and currentLV.y == desiredLV.y then
	    return 0
	end

	-- force = ( change in velocity / time ) * mass
	local mass = self.b2body:GetMass ( )
	local force = b2Vec2 ( 0, 0 )
	force.x = ( ( desiredLV.x - currentLV.x ) / seconds ) * mass
	force.y = ( ( desiredLV.y - currentLV.y ) / seconds ) * mass

	-- truncate to maximum steering force
	force:Truncate ( maxForce )
	self.b2body:ApplyForce ( force, currentPosition )

	return seconds
end

function Body:Rotate ( seconds, angle, turnRate )
	local currentAngle = self.def.angle
	local desiredAV = turnRate
	if currentAngle > angle then
		desiredAV = -desiredAV
	end

	-- condition
	local eta = self:GetTurnTime ( angle, desiredAV )
	if eta < seconds then
		self:SetAngle ( angle )
		return eta
	end

	-- output
	local step = desiredAV * seconds
	self:SetAngle ( currentAngle + step )

	-- we are rotating the body manually so clear all forces acting upon it
	self.b2body:SetAngularVelocity ( 0 )

	return seconds
end

function Body:RotateAV ( seconds, angle, turnRate )
	local currentAngle = self.def.angle
	local desiredAV = turnRate
	if currentAngle > angle then
		desiredAV = -desiredAV
	end

	-- condition
	local eta = self:GetTurnTime ( angle, desiredAV )
	if eta < seconds then
		self:SetAngle ( angle )
		return eta
	end

	-- output
	desiredAV = self:GetCorrectedAV ( desiredAV, seconds )
	self.b2body:SetAngularVelocity ( desiredAV )

	return seconds
end

function Body:RotateTorque ( seconds, angle, turnRate, maxTorque, ease )
	local currentAngle = self.def.angle
	local desiredAV = turnRate
	if currentAngle > angle then
		desiredAV = -desiredAV
	end
	local currentAV = self.b2body:GetAngularVelocity ( )

	-- condition
	if currentAV ~= 0 then
		-- eta is negative when the body is rotating in the wrong direction
		local eta = self:GetTurnTime ( angle, currentAV )
eta = math.abs ( eta )
		if eta >= 0 and eta < seconds then
			self:SetAngle ( angle )
-- todo
if ease == true then
    self.b2body:SetAngularVelocity ( 0 )
end
			return eta
		end
	else
		if angle == currentAngle then
			return 0
		end
	end

	-- output
	if ease == true then
		if currentAV ~= 0 then
			-- time required for a full stop
			local stopTime = self:GetAccelerationTimeAV ( 0, maxTorque )
			stopTime = math.abs ( stopTime )

			local eta = self:GetTurnTime ( angle, currentAV )
eta = math.abs ( eta )
			-- will we overshoot the target?
			if eta >= 0 and eta <= stopTime then

				local futureAngle = self:GetFutureAngle ( stopTime )
				-- desired velocity = ( final angle - initial angle ) / time
				desiredAV = ( angle - futureAngle ) / stopTime
				desiredAV = self:GetCorrectedAV ( desiredAV, seconds )

				-- torque = ( velocity / time ) * inertia
				local inertia = self.b2body:GetInertia ( )
				local torque = ( desiredAV / seconds ) * inertia

				-- truncate to maximum torque
				torque = math.clamp ( torque, -maxTorque, maxTorque )

				self.b2body:ApplyTorque ( torque )

				return seconds
			end
		end
	end

    self:AccelerateAV ( seconds, desiredAV, maxTorque )

	return seconds
end

function Body:AccelerateAV ( seconds, desiredAV, maxTorque )
	local currentAV = self.b2body:GetAngularVelocity ( )

	if currentAV == desiredAV then
	    return 0
	end

	-- torque = ( change in velocity / time ) * inertia
	desiredAV = self:GetCorrectedAV ( desiredAV, seconds )
	local inertia = self.b2body:GetInertia ( )
	local torque = ( ( desiredAV - currentAV ) / seconds ) * inertia

	-- truncate to maximum torque
	torque = math.clamp ( torque, -maxTorque, maxTorque )

	self.b2body:ApplyTorque ( torque )
	
	return seconds
end

function Body:FaceTorque ( seconds, position, turnRate, maxTorque, ease )
	local localPosition = self.b2body:GetLocalPoint ( position )

	-- can't face a point when we are on top of it
	if localPosition.x == 0 and localPosition.y == 0 then
		return 0
	end

	local angle = self.def.angle - localPosition:GetAngle ( )

	return self:RotateTorque ( seconds, angle, turnRate, maxTorque, ease )
end

-- keeps a dynamic body still for a desired amount of time
function Body:Stop ( delta )
    if delta then
        local av = self:GetCorrectedAV ( 0, delta )
        self.b2body:SetAngularVelocity ( av )
        local lv = self:GetCorrectedLV ( b2_b2Vec2, delta )
        self.b2body:SetLinearVelocity ( lv )
    else
        self.b2body:SetLinearVelocity ( b2_b2Vec2 )
        self.b2body:SetAngularVelocity ( 0 )
    end
end

--
-- Violation
--
function Body:Violation ( )
    --self.world:DestroyBody ( self )
end

--
-- Contact buffer
--
function Body:AddContactPoint ( body )
	-- buffer the number of contact points
	local n = self.contacts[body.name]
	if n == nil then
		self.contacts[body.name] = 1
	else
		self.contacts[body.name] = n + 1
	end
end

function Body:RemoveContactPoint ( body )
	self.contacts[body.name] = self.contacts[body.name] - 1
end

function Body:Collision ( body, position, impulseNormal )
    -- buffer all collisions that have occured within the last frame
    local collision = { body.name, position, impulseNormal }
    table.insert ( self.collisions, collision )
end

function Body:UpdateContactPoints ( seconds )
	-- degenerated bodies have no contact points
    local contacts = self.contacts
	if contacts == nil then
		return
	end
	-- remove collisions from the previous frame
	local collisions = self.collisions
	for i in pairs ( collisions ) do
        collisions[i] = nil
    end
    -- remove invalid contacts
	for i, v in pairs ( contacts ) do
		-- there was contact during the frame although the two bodies are no longer touching
		-- this occurs when the two bodies have separated or one of them was destroyed
		-- reset the contact information so that it's not reported in the following frames
		if v == 0 then
			contacts[i] = nil
		end
	end
end

function Body:IsTouching ( body )
	-- returns true if we've had mutual contact points in the last frame
	return self.contacts[body.name] ~= nil
end


--
-- Visuals
--
function Body:SetImage ( image, x, y )
    if self.image then
        self.image = nil
        self.imageX = nil
        self.imageY = nil
    end
    if image then
        self.image = image
        self.imageX = x or 0
        self.imageY = y or 0
    end
    self:Paint ( )
end
function Body:UpdateSprite ( )
	-- updates the position
	local pt = self.def.position
	self.sprite:set_position ( pt.x, pt.y )

	-- updates the rotation
	-- bodies in Box2D rotate counterclockwise
	self.sprite:set_rotation_r ( -self.def.angle )
end

function Body:Paint ( )
	local c = self.sprite.canvas
	c:clear ( )

	if self.image then
	    local image = self.world.game:FindImage ( self.image )
	    if image then
            c:set_source_image ( image, self.imageX, self.imageY )
            c:paint ( )
        else
            --self.world.game:Log ( "Could not set image '%s'", self.image )
        end
    end
    --else
if showBoxes == true then
        for i, v in pairs ( self.shapes ) do
            local type = v:GetType ( )
            if type == e_circleShape then
                c:move_to ( v.def.localPosition.x * 10, v.def.localPosition.y * 10 )
                c:circle ( v.def.radius * 10 )
            elseif type == e_polygonShape then
                local vertices = v.vertices
                c:move_to ( vertices[1].x * 10, vertices[1].y * 10 )
                for i = 2, #vertices do
                    c:line_to ( vertices[i].x * 10, vertices[i].y * 10 )
                end
                c:close_path ( )
--[[
                local glow = game:DemandImage ( "Art/glow.png" )
                local last = vertices[#vertices]
                for i = 1, #vertices do
                    local d = last:Distance ( vertices[i] )
                    local steps = d / 5
                    local sx = ( vertices[i].x - last.x ) / steps
                    local sy = ( vertices[i].y - last.y ) / steps
                    for i = 1, steps, 1 do
                        c:set_source_image ( glow, last.x + sx * i, last.y + sy * i )
                        c:paint ( )
                    end
                    last = vertices[i]
                end
                ]]
            end

            local a = 1
            if v.def.isSensor == true then
                a = 0.25
            end

            local color = WHITE
            if v.def.density > 0 then
                color = GREEN
            end

            --c:set_fill_style ( SILVER, a )
            --c:fill_preserve ( )
--[[
            c:set_line_style ( 10, YELLOW, 0.1 )
            c:stroke_preserve ( )
            c:set_line_style ( 8, YELLOW, 0.1 )
            c:stroke_preserve ( )
            c:set_line_style ( 6, YELLOW, 0.1 )
            c:stroke_preserve ( )
            c:set_line_style ( 4, YELLOW, 0.1 )
            c:stroke_preserve ( )
            c:set_line_style ( 2, YELLOW, 0.1 )
            c:stroke_preserve ( )
            c:set_line_style ( 1, YELLOW, 0.1 )
            c:stroke_preserve ( )
]]

            c:set_line_style ( 1, RED, 0.5 )
            c:stroke ( )
end
        end
    --end
end

--
-- Goals
--
function Body:UpdateGoal ( seconds )
	if self.goal then
		local status = self.goal:Process ( seconds, self )
		if status ~= GOAL_ACTIVE then
		    self.goal:Destroy ( )
			self.goal = nil
		end
	end
end

function Body:SetGoal ( goal )
	if self.goal then
        self.goal:Destroy ( )
	end
	self.goal = goal
end


--
-- Messaging
--
function Body:Message ( type, ... )
	if self.goal then
		return self.goal:Message ( self, type, ... )
	end
	return false
end

--
-- XML
--
function Body:CreateShapeXML ( shape )
	local density = tonumber ( shape.density )
	local friction = tonumber ( shape.friction )
	local restitution = tonumber ( shape.restitution )
	local isSensor = toboolean ( shape.isSensor )

	local s
	if shape.type == 'circle' then
		local radius = tonumber ( xml.find ( shape, 'Radius' )[1] )
		local localPosition = tob2Vec2 ( xml.find ( shape, 'LocalPosition' )[1] )

		s = self:CreateShapeCircle ( radius, localPosition, density, friction, restitution, isSensor )
	elseif shape.type == 'polygon' then
		local path = xml.find ( shape, 'Path' )
		-- polygon shape without a path
		assert ( path, "Polygon shape without a path" )
		local vertices = toPath ( path )

		s = self:CreateShapePolygon ( vertices, density, friction, restitution, isSensor )
	else
		-- unknown shape type
		assert ( false, "Unknown shape type" )
	end

	-- could not generate shape
--	assert ( s, "Could not generate shape" )

	local filter = xml.find ( shape, 'Filter' )
	if filter then
		local groupIndex = tonumber ( xml.find ( filter, 'GroupIndex' )[1] )
		local categoryBits = tonumber ( xml.find ( filter, 'CategoryBits' )[1] )
		local maskBits = tonumber ( xml.find ( filter, 'MaskBits' )[1] )
		--s:SetFilter ( groupIndex, categoryBits, maskBits )
	end
    
	return s
end

function Body:GetXML ( )
	local body = xml.new ( "Body" )

	body.name = self.name
	body.position = b2Vec2tostring ( self.def.position )
	body.angle = self.def.angle
	body.linearDamping = self.def.linearDamping
	body.angularDamping = self.def.angularDamping
	body.fixedRotation = self.def.fixedRotation
	body.isBullet = self.def.isBullet
	body.allowSleep = self.def.allowSleep
	body.isSleeping = self.def.isSleeping

	if self.scale ~= 1 then
		body:append ( "Scale" )[1] = self.scale
	end

	if self.b2body then
		local lv = self.b2body:GetLinearVelocity ( )
		if lv.x ~= 0 or lv.y ~= 0 then
			body:append ( "LinearVelocity" )[1] = b2Vec2tostring ( lv )
		end
		local av = self.b2body:GetAngularVelocity ( )
		if av ~= 0 then
			body:append ( "AngularVelocity" )[1] = av
		end
	end

	for i, v in pairs ( self.shapes ) do
		local shape = v:GetXML ( )
		body:append ( shape )
	end

	if self.goal then
		local goal = self.goal:GetXML ( )
		body:append ( goal )
	end

	local sprite = body:append ( "Sprite" )
	sprite.alpha = self.sprite.alpha
	sprite.depth = self.sprite.depth
	sprite.color = Colortostring ( self.sprite.color )
	sprite.visible = self.sprite.visible
	if self.image then
    	local image = sprite:append ( "Image" )
        image[1] = self.image
        image.x = self.imageX
        image.y = self.imageY
    end

	return body
end
