--
-- Game
--
Game = {}
GameMT = { __index = Game }

function Game:Create ( title, width, height, depth, windowed )
	local self = {}
	setmetatable ( self, GameMT )

    math.randomseed ( os.time ( ) )
    
	display:create ( title, width, height, depth, windowed )

	local camera = Camera ( )
	local scene = Layer ( )
	scene:add_child ( camera )
	
	local viewport = Viewport ( 0, 0, width, height )
	viewport.width = width
	viewport.height = height
	viewport.camera = camera
	viewport.bgcolor = BLACK
	display.viewport:add_child ( viewport )

	if audio then
		audio:set_mode ( 16, 2, 44100 )
		self.audio = audio
	end

	self.camera = camera
	self.scene = scene
	self.viewport = viewport
	self.streams = {}
	self.accumulator = 0
	self.timeFactor = 1
	self.elapsed = 0
	self.state = 'inactive'

	self.timer = Timer ( )
    self.timer.on_tick = function ( timer )
        local seconds = timer:get_delta_ms ( ) / 1000 * self.timeFactor
        self:Update ( timer.interval / 1000, seconds )
    end
    self.timer:start ( 16, true )

	self.images = {}
	self.sounds = {}

    self.log = {}
	self.log_file = io.open ( "game.log", "w+" )

    self.font = Font ( )
    --self.font:load_system ( "Courier New", 8 )
    --self.font:load_file ( "Fonts/franchise.ttf", 10 )
	--self.font:load_file ( "Fonts/uni05_53.ttf", 12 )
	--self.font:load_file ( "Fonts/visitor1.ttf", 14 )
	self.font:load_file ( "Fonts/JOYSTIX.TTF", 12 )
	self.terminal = Sprite ( )
	self.terminal.canvas:set_font ( self.font, WHITE, 0.75 )
	self.viewport:add_child ( self.terminal )

	self.query = nil

	self:Log ( "Creating game\n" )

	return self
end

function Game:Destroy ( )
	self:Log ( "Destroying game\n" )

	display:destroy ( )

	self.camera = nil
	self.scene = nil
	self.streams = nil
	--self.player = nil
	self.accumulator = nil
	self.timeFactor = nil
	self.elapsed = nil
	self.state = nil
	
	--self.gui:Destroy ( )
	--self.gui = nil
	
	self.timer:stop ( )
	self.timer = nil
	
	self.images = nil
	self.sounds = nil

	self.log = nil
	self.log_file:close ( )
	self.log_file = nil

    self.font:unload ( )
	self.font = nil
	self.terminal = nil
	
	self.query = nil
	
    self:StopMusic ( )
end

function Game:Log ( sz, ... )
	if self.log then
		sz = string.format ( sz, ... )
		table.insert ( self.log, sz )

		self.log_file:write ( sz )
		self.log_file:flush ( )
	end
end

--
-- Image management
--
function Game:DemandImageDirectory ( path, recursive )
    local fullPath = lfs.currentdir ( ) .. '/' .. path
	for file in lfs.dir ( fullPath ) do
		if file ~= '.' and file ~= '..' then
			local attr = lfs.attributes ( fullPath .. '/' .. file )
			if attr.mode == 'directory' then
				if recursive == true then
				    self:DemandImageDirectory ( path .. '/' .. file, recursive )
				end
			else
			    self:DemandImage ( path .. '/' .. file )
			end
		end
	end
end

function Game:DemandImage ( name )
    local image = self.images[name]
    if image == nil then
        image = Image ( )
        if image:load ( name ) == false then
            self:Log ( "Could not load image '%s'", name )
        end
        self.images[name] = image
        image.refCount = 0
    end
    image.refCount = image.refCount + 1
    return image
end

function Game:UnDemandImage ( name )
    local image = self.images[name]
    image.refCount = image.refCount - 1
    if image.refCount == 0 then
        image.refCount = nil
        image:unload ( )
        self.images[name] = nil
    end
end

function Game:FindImage ( name )
    return self.images[name]
end

--
-- Sound management
--
function Game:DemandSoundDirectory ( path, recursive )
    local fullPath = lfs.currentdir ( ) .. '/' .. path
	for file in lfs.dir ( fullPath ) do
		if file ~= '.' and file ~= '..' then
			local attr = lfs.attributes ( fullPath .. '/' .. file )
			if attr.mode == 'directory' then
				if recursive == true then
				    self:DemandSoundDirectory ( path .. '/' .. file, recursive )
				end
			else
			    self:DemandSound ( path .. '/' .. file )
			end
		end
	end
end

function Game:DemandSound ( name )
    local sound = self.sounds[name]
    if sound == nil then
        sound = Sound ( )
        if sound:load ( name ) == false then
            self:Log ( "Could not load sound '%s'", name )
        end
        self.sounds[name] = sound
        sound.refCount = 0
    end
    sound.refCount = sound.refCount + 1
    return sound
end

function Game:UnDemandSound ( name )
    local sound = self.sounds[name]
    sound.refCount = sound.refCount - 1
    if sound.refCount == 0 then
        sound.refCount = nil
        sound:unload ( )
        self.sounds[name] = nil
    end
end

function Game:FindSound ( name )
    return self.sounds[name]
end

function Game:PlayMusic ( name, loop, volume )
    if self.music then
        self.music:stop ( )
    end
    local m = Music ( name )
    if loop == true then
        self.music = self.audio:play_loop ( m )
    else
        self.music = self.audio:play ( m )
    end
    self.music.volume = volume or 1
end

function Game:StopMusic ( )
    if self.music then
        self.music:stop ( )
        self.music = nil
    end
end


--
-- Stream management
--
function Game:CreateStream ( file )
	-- file is already being streamed
	assert ( self.streams[file] == nil, "File is already streamed" )

	local stream = Stream:Create ( self, file )
	self.streams[file] = stream

	return stream
end

function Game:DestroyStream ( stream )
	stream:Destroy ( )

	self.streams[stream.name] = nil
end

function Game:FindStream ( file )
	return self.streams[file]
end

function Game:FindStreamFile ( world )
	assert ( world, "World parameter is nil" )
	for i, v in pairs ( self.streams ) do
		if v.world == world then
			return i
		end
	end
	return nil
end

--
-- Update
--
function Game:Update ( interval, delta )
    -- update terminal
    local width = self.viewport.width
    local height = self.viewport.height
    local lines = math.floor ( height / game.font:get_size ( ) )

    local c = self.terminal.canvas
    c:clear ( )
    if self.query then
        local body
        for i, v in pairs ( self.streams ) do
            if v.selects > 0 then
                body = v.world:FindBody ( self.query )
                if body then
                    break
                end
            end
        end
        --local body = self.player.world:FindBody ( self.query )
        if body then
            local q = body:GetXML ( )
            c:move_to ( -width / 2, height / 2 - 12 )
            local sz = xml.str ( q )
            sz = string.sub ( sz, 0, lines * ( width / 12 ) )
            c:write ( sz )
        end
    else
    --[[
        local current = math.max ( #self.log - lines, 1 )
        for i = 1, math.min ( #self.log, lines ), 1 do
            c:move_to ( -width / 2, height / 2 - i * game.font:get_size ( ) )
            c:write ( self.log[current] )
            current = current + 1
        end
        ]]
    end

	if self.state ~= 'playing' then
		return
	end

    self.elapsed = self.elapsed + delta
        
	--collectgarbage ( "stop" )

	for i, v in pairs ( self.streams ) do
		v:Update ( )
	end

	-- update with accumulator
	self.accumulator = self.accumulator + delta
	while self.accumulator > interval do
		for i, v in pairs ( self.streams ) do
			v:Simulate ( interval )
		end
		self.accumulator = self.accumulator - interval 
	end
--[[
	-- update without accumulator
	for i, v in pairs ( self.streams ) do
		v:Simulate ( delta )
	end
]]
	-- sync the sprites
	for i, v in pairs ( self.streams ) do
		v:Syncronize ( delta )
	end

	--collectgarbage ( "collect" )
end

function Game:SetTimeFactor ( factor )
    self.timeFactor = factor
end

function Game:Pause ( )
    self:Log ( "Game paused" )
    self.state = 'paused'
    if self.music then
        self.music:pause ( )
    end
end

function Game:Play ( )
    self:Log ( "Game resumed" )
    self.state = 'playing'
    if self.music then
        self.music:resume ( )
    end
end