--
-- Shape
--
Shape = {}
ShapeMT = { __index = Shape }

ShapeCircle = {}
ShapeCircleMT = { __index = ShapeCircle }

setmetatable ( ShapeCircle, { __index = Shape } )

function ShapeCircle:Create ( body, radius, localPosition, density, friction, restitution, isSensor )
	local shape = {}
	setmetatable ( shape, ShapeCircleMT )

	shape.body = body

	local def = b2CircleDef ( )
	def.radius = radius
	def.localPosition = localPosition
	-- physical properties
	def.density = density
	def.friction = friction
	def.restitution = restitution
	def.isSensor = isSensor

	shape.def = def

	if shape:Generate ( ) == false then
		return nil
	end

	return shape
end

function ShapeCircle:Destroy ( )
	self:Degenerate ( )

	self.body = nil
	self.def = nil
end

function ShapeCircle:Generate ( )
	self:Degenerate ( )

    local proxies = self.body.world.b2world:GetProxyCount ( )
	assert ( proxies < b2_maxProxies, "b2_maxProxies exceeded" )

	if self.def.radius == 0 then
		return false
	end

	local c = self.body.b2body:CreateShape ( self.def )

	if c == nil then
		return false
	end

	c.parent = self
	self.b2shape = c

	self.body.b2body:SetMassFromShapes ( )

	return true
end

function ShapeCircle:Degenerate ( )
	if self.b2shape == nil then
		return
	end

	local body = self.body.b2body
	body:DestroyShape ( self.b2shape )
	self.b2shape.parent = nil

	self.b2shape = nil
end

function ShapeCircle:GetType ( )
	return e_circleShape
end

function ShapeCircle:GetBody ( )
	return self.body
end

function ShapeCircle:TestPoint ( xf, pt )
	return self.b2shape:TestPoint ( xf, pt )
end

function ShapeCircle:SetFilter ( groupIndex, categoryBits, maskBits )
	local filter = self.def.filter
	filter.groupIndex = groupIndex
	filter.categoryBits = categoryBits
	filter.maskBits = maskBits

	self.b2shape:SetFilterData ( filter )

	self.body.world.b2world:Refilter ( self.b2shape )
end


ShapePolygon = {}
ShapePolygonMT = { __index = ShapePolygon }

setmetatable ( ShapePolygon, { __index = Shape } )

function ShapePolygon:Create ( body, vertices, density, friction, restitution, isSensor )
	local shape = {}
	setmetatable ( shape, ShapePolygonMT )

	shape.body = body

    -- shape path
	shape.vertices = vertices
	shape.triangles = {}

	local def = b2PolygonDef ( )
	-- physical properties
	def.density = density
	def.friction = friction
	def.restitution = restitution
	def.isSensor = isSensor

	shape.def = def

	if shape:Generate ( ) == false then
		return nil
	end

	return shape
end

function ShapePolygon:Destroy ( )
	self:Degenerate ( )

	self.body = nil
	self.def = nil
	self.vertices = nil
	self.triangles = nil
end

function ShapePolygon:Generate ( )
	self:Degenerate ( )

	local out = {}
	local ret = b2Fill ( self.vertices, out )
	if ret == false then
		return false
	end

	local b2body = self.body.b2body
	local def = self.def

	def.vertexCount = 3
	for i = 1, #out, 3 do
		def.vertices[0]:Set ( out[i].x, out[i].y )
		def.vertices[1]:Set ( out[i + 1].x, out[i + 1].y )
		def.vertices[2]:Set ( out[i + 2].x, out[i + 2].y )

        local proxies = self.body.world.b2world:GetProxyCount ( )
		assert ( proxies < b2_maxProxies, "b2_maxProxies exceeded" )

		local t = b2body:CreateShape ( def )
		if t == nil then
			-- polygon was partially generated
			self:Degenerate ( )
			return false
		end

		t.parent = self
		table.insert ( self.triangles, t )
	end

	b2body:SetMassFromShapes ( )

	return true
end

function ShapePolygon:Degenerate ( )
	local b2body = self.body.b2body
	while #self.triangles > 0 do
		local last = table.remove ( self.triangles )
		b2body:DestroyShape ( last )
		last.parent = nil
	end
end

function ShapePolygon:GetType ( )
	return e_polygonShape
end

function ShapePolygon:GetBody ( )
	return self.body
end

function ShapePolygon:TestPoint ( xf, pt )
	for i, v in pairs ( self.triangles ) do
		if v:TestPoint ( xf, pt ) == true then
			return true
		end
	end
	return false
end

function ShapePolygon:SetFilter ( groupIndex, categoryBits, maskBits )
	local filter = self.def.filter
	filter.groupIndex = groupIndex
	filter.categoryBits = categoryBits
	filter.maskBits = maskBits

	for i, v in pairs ( self.triangles ) do
		v:SetFilterData ( filter )
		self.body.world.b2world:Refilter ( v )
	end
end

--
-- XML
--
function Shape:GetXML ( )
	local shape = xml.new ( "Shape" )

	local type = self:GetType ( )

	shape.density = self.def.density
	shape.friction = self.def.friction
	shape.restitution = self.def.restitution
	shape.isSensor = self.def.isSensor

	if type == e_circleShape then
	    shape.type = 'circle'
		shape:append ( "Radius" )[1] = self.def.radius
		shape:append ( "LocalPosition" )[1] = b2Vec2tostring ( self.def.localPosition )
	elseif type == e_polygonShape then
	    shape.type = 'polygon'
	    shape.path = Pathtostring ( self.vertices )
	end

	local filter = shape:append ( "Filter" )
	filter:append ( "GroupIndex" )[1] = self.def.filter.groupIndex
	filter:append ( "CategoryBits" )[1] = self.def.filter.categoryBits
	filter:append ( "MaskBits" )[1] = self.def.filter.maskBits

	return shape
end

