--
-- Stream
--
Stream = {}
StreamMT = { __index = Stream }

function Stream:Create ( game, file )
	local stream = {}
	setmetatable ( stream, StreamMT )

	stream.game = game
	stream.file = file
	stream.state = 'unloaded'
	-- demanded streams are scheduled for loading
	stream.demands = 0
	-- selected streams are loaded right away
	stream.selects = 0
	-- links to adjacent streams
	stream.links = {}

	return stream
end

function Stream:Destroy ( )
	self:UnLoad ( )

	self.game = nil
	self.file = nil
	self.state = nil
	self.demands = nil
	self.selects = nil
	self.links = nil
end

function Stream:Load ( isCoroutine )
    -- creates world from an xml file right away or as a coroutine
	self.world = World:CreateXML ( self.game, self.file, isCoroutine )

	self.state = 'loaded'

	if self.OnLoad then
		local res, msg = pcall ( self.OnLoad, self )
		if res == false then
		 	self.game:Log ( 'OnLoad failed in stream "%s"\n', self.file )
		 	self.game:Log ( msg .. '\n' )
		end
	end

	self.game:Log ( 'Loaded stream "%s"\n', self.file )
end

function Stream:UnLoad ( isCoroutine )
	if self.OnUnLoad then
		self:OnUnLoad ( )
	end

    -- destroys world right away or as a coroutine
	self.world:Destroy ( isCoroutine )
	self.world = nil
	self.state = 'unloaded'

	self.game:Log ( 'Unloaded stream "%s"\n', self.file )
end

function Stream:UpdateCoroutine ( complete )
    -- complete outstanding loading or unloading coroutines
	local status
	while true do
		coroutine.resume ( self.co, self, true )
		status = coroutine.status ( self.co )
		if complete == false or status == 'dead' then
			break
		end
	end

	if status == 'dead' then
		-- make sure the loading or unloading coroutine has not failed
		if self.state == 'loading' or self.state == 'unloading' then
			self.game:Log ( 'Coroutine failed in stream "%s" while "%s"\n', self.file, self.state )
		end

		self.co = nil
	end
end

function Stream:Update ( delta )
	-- are we currently loading or unloading the world?
	if self.co then
		self:UpdateCoroutine ( false )
	end
end

function Stream:Simulate ( delta )
    -- updates the physical propereties of the loaded world
	if self.state == 'loaded' and self.selects > 0 then
		self.world:Step ( delta, 10 )

        if self.OnSimulate then
            local res, msg = pcall ( self.OnSimulate, self, delta )
            if res == false then
                self.game:Log ( 'OnSimulate failed in stream "%s"\n', self.file )
                self.game:Log ( msg .. '\n' )
            end
        end
	end
end

function Stream:Syncronize ( delta )
    -- syncronizes the visuals to the physical world
	if self.state == 'loaded' and self.selects > 0 then
		self.world:Syncronize ( delta )
	end
end

function Stream:Demand ( complete )
	self.demands = self.demands + 1

	-- already loaded
	if self.state == 'loaded' then
		return
	end

	if self.state == 'unloading' then
		-- complete outstanding coroutine
		self:UpdateCoroutine ( true )
		assert ( self.state == 'unloaded', "Unloading stream failed" )
	end

	if complete == true then
		if self.state == 'unloaded' then
			self:Load ( )
		elseif self.state == 'loading' then
			-- complete outstanding coroutine
			self:UpdateCoroutine ( true )
			assert ( self.state == 'loaded', "Loading stream failed" )
		end
	else
		if self.state == 'unloaded' then
			self.co = coroutine.create ( self.Load )
			self.state = 'loading'
		end
	end
end

function Stream:UnDemand ( complete )
	self.demands = self.demands - 1

	-- bad reference counting
	assert ( self.demands >= 0, "Bad reference counting" )

	-- don't unload if there are demands for this stream
	if self.demands > 0 then
		return
	end

	-- already unloaded
	if self.state == 'unloaded' then
		return
	end

	if self.state == 'loading' then
		-- complete outstanding coroutine
		self:UpdateCoroutine ( true )
		assert ( self.state == 'loaded', "Loading stream failed" )
	end

	if complete == true then
		if self.state == 'loaded' then
			self:UnLoad ( )
		end
	else
		if self.state == 'loaded' then
			self.co = coroutine.create ( self.UnLoad )
			self.state = 'unloading'
		end
	end
end

function Stream:LinkStream ( stream )
	self.links[stream] = true
end

function Stream:UnLinkStream ( stream )
	self.links[stream] = nil
end

function Stream:Select ( )
	self.selects = self.selects + 1

	if self.selects ~= 1 then
		return
	end

	self:Demand ( true )
	-- locate and demand adjacent streams
	for i in pairs ( self.links ) do
		local stream = self.game:FindStream ( i )
		stream:Demand ( false )
	end

	self.game:Log ( "Selected stream '%s'\n", self.file )
end

function Stream:DeSelect ( forced )
	self.selects = self.selects - 1

	-- bad reference counting
	assert ( self.selects >= 0, "Bad reference counting" )

	if self.selects ~= 0 then
		return
	end

	self:UnDemand ( forced or false )
	-- locate and un-demand adjacent streams
	for i in pairs ( self.links ) do
		local stream = self.game:FindStream ( i )
		stream:UnDemand ( forced or false )
	end

	self.game:Log ( "Deselected stream '%s'\n", self.file )
end
