--
-- World
--
World = {}
WorldMT = { __index = World }

function World:Create ( game, left, top, right, bottom, gravity, doSleep )
	local world = {}
	setmetatable ( world, WorldMT )

	world.game = game

	-- AABB has to be properly aligned
	assert ( left < right, "Improperly aligned AABB" )
	assert ( top < bottom, "Improperly aligned AABB" )

	world.aabb = b2AABB ( )
	world.aabb.lowerBound:Set ( left, top )
	world.aabb.upperBound:Set ( right, bottom )
	world.bodies = {}
	world.joints = {}
	world.gravity = gravity
	world.doSleep = doSleep
	world.nextID = 0

	world.sprite = Sprite ( )
	world.layer = Layer ( )

	if world:Generate ( ) == false then
		return nil
	end

	world:Paint ( )

	return world
end

function World:Destroy ( isCoroutine )
	self:DestroyAllJoints ( isCoroutine )
	self:DestroyAllBodies ( isCoroutine )

	self:Degenerate ( )

	self.sprite = nil
	self.layer = nil

	self.aabb = nil
	self.bodies = nil
	self.joints = nil
	self.gravity = nil
	self.doSleep = nil
	self.nextID = nil

	self.game = nil
end

function World:Generate ( )
	self:Degenerate ( )

	self.b2world = b2World ( self.aabb, self.gravity, self.doSleep )
	if self.b2world == nil then
		return false
	end
	self.b2world.parent = self

	self.layer:add_child ( self.sprite )
	self.game.scene:add_child ( self.layer )

    -- boundary listener
	local listener = b2BoundaryListener ( )
	listener.Violation = function ( listener, body )
		self:Violation ( body.parent )
	end
	self.b2world:SetBoundaryListener ( listener )

	-- contact listeners and filters must be set before any bodies are generated
	local listener = b2ContactListener ( )
	listener.Add = function ( listener, cp )
		self:AddContactPoint ( cp )
	end
	listener.Remove = function ( listener, cp )
		self:RemoveContactPoint ( cp )
	end
	listener.Result = function ( listener, cr )
	    self:Collision ( cr )
	end
	self.b2world:SetContactListener ( listener )

	for i, v in pairs ( self.bodies ) do
		v:Generate ( )
	end
	for i, v in pairs ( self.joints ) do
		v:Generate ( )
	end
end

function World:Degenerate ( )
	if self.b2world == nil then
		return
	end

	for i, v in pairs ( self.joints ) do
		v:Degenerate ( )
	end
	for i, v in pairs ( self.bodies ) do
		v:Degenerate ( )
	end

	self.b2world:SetContactListener ( nil )

	self.b2world:SetBoundaryListener ( nil )

	self.game.scene:remove_child ( self.layer )

	self.b2world.parent = nil
	self.b2world = nil
end

function World:SetGravity ( gravity )
	self.gravity:Set ( gravity.x, gravity.y )
	self.b2world:SetGravity ( gravity )
end

function World:GetGravity ( )
	return self.gravity
end

--
-- Bodies
--
function World:GetAvailableName ( )
	local name
	while true do
		name = "body" .. self.nextID
		if self.bodies[name] == nil then
			break
		end
		self.nextID = self.nextID + 1
	end

	return name
end

function World:FindBody ( name )
	return self.bodies[name]
end

function World:RenameBody ( old, new )
	local bodies = self.bodies
	if bodies[new] ~= nil then
		return false
	end
	if bodies[old] == nil then
		return false
	end
	bodies[new] = bodies[old]
	bodies[old] = nil
	bodies[new].name = new

	return true
end

function World:CreateBody ( name, position, angle, linearDamping, angularDamping, fixedRotation, isBullet, allowSleep, isSleeping )
	-- name is already registered
	if self.bodies[name] ~= nil then
		return nil
	end

	local body = Body:Create ( self, position, angle, linearDamping, angularDamping, fixedRotation, isBullet, allowSleep, isSleeping )

	if body then
		body.name = name
		self.bodies[name] = body
	end

	return body
end

function World:DuplicateBody ( body, name )
	-- notice that the source body doesn't need to be from the same world
	local position = b2Vec2 ( body.def.position.x, body.def.position.y )
	local b = self:CreateBody ( name, position, body.def.angle, body.def.linearDamping, body.def.angularDamping, body.def.fixedRotation, body.def.isBullet, body.def.allowSleep, body.def.isSleeping )

	if b then
		-- duplicate all shapes
		for i, v in pairs ( body.shapes ) do
			local type = v:GetType ( )
			local s
			if type == e_circleShape then
				local localPosition = b2Vec2 ( v.def.localPosition.x, v.def.localPosition.y )
				s = b:CreateShapeCircle ( v.def.radius, localPosition, v.def.density, v.def.friction, v.def.restitution, v.def.isSensor )
			elseif type == e_polygonShape then
				local vertices = {}
				for i, v in pairs ( v.def.vertices ) do
					table.insert ( vertices, v )
				end
				s = b:CreateShapePolygon ( vertices, v.def.density, v.def.friction, v.def.restitution, v.def.isSensor )
			end
			s:SetFilter ( v.def.filter.groupIndex, v.def.filter.categoryBits, v.def.filter.maskBits )
		end
		b:SetImage ( body.image, body.imageX, body.imageY )
	end

	return b
end

function World:DestroyBody ( body )
	body:Destroy ( )

	self.bodies[body.name] = nil
	body.name = nil
end

function World:DestroyAllBodies ( isCoroutine )
	for i, v in pairs ( self.bodies ) do
		self:DestroyBody ( v )
		if isCoroutine == true then
			coroutine.yield ( )
		end
	end
end

--
-- Joints
--
function World:CreateJointDistance ( body1, body2, collideConnected, a1, a2 )
	local joint = DistanceJoint:Create ( self, body1, body2, collideConnected, a1, a2 )
	if joint then
		table.insert ( self.joints, joint )
	end

	return joint
end

function World:CreateJointPulley ( body1, body2, collideConnected, ga1, ga2, a1, a2, ratio )
	local joint = PulleyJoint:Create ( self, body1, body2, collideConnected, ga1, ga2, a1, a2, ratio )
	if joint then
		table.insert ( self.joints, joint )
	end

	return joint
end

function World:CreateJointRevolute ( body1, body2, collideConnected, a )
	local joint = RevoluteJoint:Create ( self, body1, body2, collideConnected, a )
	if joint then
		table.insert ( self.joints, joint )
	end

	return joint
end

function World:CreateJointPrismatic ( body1, body2, collideConnected, a, axis )
	local joint = PrismaticJoint:Create ( self, body1, body2, collideConnected, a, axis )
	if joint then
		table.insert ( self.joints, joint )
	end

	return joint
end
--[[
function World:CreateJointGear ( body1, body2, collideConnected, a, axis )
	local joint = PrismaticJoint:Create ( self, body1, body2, collideConnected, a, axis )
	if joint then
		table.insert ( self.joints, joint )
	end

	return joint
end
]]
function World:DestroyJoint ( joint )
	joint:Destroy ( )

	for i, v in pairs ( self.joints ) do
		if v == joint then
			table.remove ( self.joints, i )
			break
		end
	end
end

function World:DestroyAllJoints ( isCoroutine )
	while #self.joints > 0 do
		local last = table.remove ( self.joints )
		last:Destroy ( )
		if isCoroutine == true then
			coroutine.yield ( )
		end
	end
end

function World:GetAssociatedJoints ( body )
	local result = {}
	for i, v in pairs ( self.joints ) do
		if v.body1 == body or v.body2 == body then
			table.insert ( result, v )
		end
	end
	return result
end

function World:Step ( seconds, iterations )
	-- update local information
	for i, v in pairs ( self.bodies ) do
		v:Update ( )
	end
	-- process goals
	for i, v in pairs ( self.bodies ) do
		v:UpdateGoal ( seconds )
	end
	-- flush contact buffers
	for i, v in pairs ( self.bodies ) do
		v:UpdateContactPoints ( seconds )
	end

	-- simulate the physical world
	self.b2world:Step ( seconds, iterations )
end

function World:Syncronize ( seconds )
    -- syncronize sprites to bodies
	for i, v in pairs ( self.bodies ) do
		v:UpdateSprite ( )
	end
end

--
-- Query functions
--
function World:QueryPoint ( pt, limit )
	local aabb = b2AABB ( )
	aabb.lowerBound:Set ( pt.x - 1, pt.y - 1 )
	aabb.upperBound:Set ( pt.x + 1, pt.y + 1 )

	local q = {}
	local proxies = self.b2world:GetProxyCount ( )
	self.b2world:Query ( aabb, q, proxies )

	local array = {}
	for i, v in pairs ( q ) do
		-- filter duplicate shapes
		local exists = false
		for i2, v2 in pairs ( array ) do
			if v2 == v.parent then
				exists = true
				break
			end
		end

		if exists == false then
			local b2body = v:GetBody ( )
			local xform = b2body:GetXForm ( )
			if v:TestPoint ( xform, pt ) == true then
				table.insert ( array, v.parent )
				if limit and limit == #array then
					break
				end
			end
		end
	end

	return array
end

function World:QueryAABB ( left, top, right, bottom, limit )
	-- AABB has to be properly aligned
	assert ( left < right, "Improperly aligned AABB" )
	assert ( top < bottom, "Improperly aligned AABB" )

	local aabb = b2AABB ( )
	aabb.lowerBound:Set ( left, top )
	aabb.upperBound:Set ( right, bottom )

	local q = {}
	local proxies = self.b2world:GetProxyCount ( )
	self.b2world:Query ( aabb, q, proxies )

	local array = {}
	for i, v in pairs ( q ) do
		-- filter duplicate shapes
		local exists = false
		for i2, v2 in pairs ( array ) do
			if v2 == v.parent then
				exists = true
				break
			end
		end

		if exists == false then
			table.insert ( array, v.parent )
			if limit and limit == #array then
				break
			end
		end
	end

	return array
end

--
-- Visuals
--
function World:Paint ( )
	local c = self.sprite.canvas

	c:clear ( )

	local l = self.aabb.lowerBound.x
	local t = self.aabb.lowerBound.y
	local r = self.aabb.upperBound.x
	local b = self.aabb.upperBound.y
	c:move_to ( l, t )
	c:line_to ( l, b )
	c:line_to ( r, b )
	c:line_to ( r, t )
	c:close_path ( )

    --c:set_line_style ( 1, GRAY, 1 )
	--c:stroke ( )
	c:set_fill_style ( PINK, 0.1 )
	--c:fill ( )
end

--
-- Boundary listener
--
function World:Violation ( body )
    body:Violation ( )
end

--
-- Contact listeners and filters
--
function World:AddContactPoint ( cp )
	local body1 = self.bodies[cp.shape1.parent.body.name]
	local body2 = self.bodies[cp.shape2.parent.body.name]

	body1:AddContactPoint ( body2 )
	body2:AddContactPoint ( body1 )
end

function World:RemoveContactPoint ( cp )
	local body1 = self.bodies[cp.shape1.parent.body.name]
	local body2 = self.bodies[cp.shape2.parent.body.name]

	body1:RemoveContactPoint ( body2 )
	body2:RemoveContactPoint ( body1 )
end

function World:Collision ( cr )
    local body1 = self.bodies[cr.shape1.parent.body.name]
    local body2 = self.bodies[cr.shape2.parent.body.name]

    --if cr.shape1:IsSensor ( ) == false and cr.shape2:IsSensor ( ) == false then
        body1:Collision ( body2, b2Vec2 ( cr.position ), cr.normalImpulse )
        body2:Collision ( body1, b2Vec2 ( cr.position ), cr.normalImpulse )
    --end
end

--
-- XML
--
function World:CreateXML ( game, file, isCoroutine )
	local w

	local xmlbox2d = xml.load ( file )
	-- could not load XML file
	assert ( xmlbox2d, "Could not load XML file" )

	-- file is not XMLBox2D format
	assert ( xml.tag ( xmlbox2d ) == "XMLBox2D", "File is not XMLBox2D format" )

	-- incorrect file version
	assert ( xmlbox2d.version == "1.0", "Incorrect file version" )

	for i = 1, #xmlbox2d, 1 do
		local world = xmlbox2d[i]
		-- unknown tag
		assert ( xml.tag ( world ) == 'World', "Unkown tag" )

		local left = tonumber ( world.left )
		local top = tonumber ( world.top )
		local right = tonumber ( world.right )
		local bottom = tonumber ( world.bottom )

		local gravity = tob2Vec2 ( world.gravity )
		local doSleep = toboolean ( world.doSleep )

		w = World:Create ( game, left, top, right, bottom, gravity, doSleep )

		-- could not generate world
		assert ( w, "Could not generate world" )

		local objects = #world
		for i = 1, objects, 1 do
			local object = world[i]
			local tag = xml.tag ( object )

			if tag == 'Body' then
				w:CreateBodyXML ( object )
			elseif tag == 'Joint' then
				w:CreateJointXML ( object )
			else
				-- unknown tag in world
				assert ( false, "Unkown tag in world" )
			end

			if isCoroutine == true then
				coroutine.yield ( )
			end
		end
	end

	return w
end

function World:CreateJointXML ( joint )
	local body1 = self:FindBody ( joint.body1 )
	local body2 = self:FindBody ( joint.body2 )

	local anchor1 = xml.find ( joint, 'Anchor1' )
	local anchor2 = xml.find ( joint, 'Anchor2' )
	local collideConnected = toboolean ( joint.collideConnected )

	local j
	if joint.type == 'distance' then
		local a1 = tob2Vec2 ( anchor1[1] )
		local a2 = tob2Vec2 ( anchor2[1] )

		j = self:CreateJointDistance ( body1, body2, collideConnected, a1, a2 )

		local frequencyHz = xml.find ( joint, 'FrequencyHz' )
		if frequencyHz then
			j.def.frequencyHz = tonumber ( frequencyHz[1] )
		end

		local dampingRatio = xml.find ( joint, 'DampingRatio' )
		if dampingRatio then
			j.def.dampingRatio = tonumber ( dampingRatio[1] )
		end

		j:Generate ( )
	elseif joint.type == 'revolute' then
		local a = tob2Vec2 ( anchor1[1] )

		j = self:CreateJointRevolute ( body1, body2, collideConnected, a )

		local limit = xml.find ( joint, 'Limit' )
		if limit then
			j:SetLimit ( toboolean ( limit.enabled ), tonumber ( limit.lowerAngle ), tonumber ( limit.upperAngle ) )
		end
		local motor = xml.find ( joint, 'Motor' )
		if motor then
			j:SetMotor ( toboolean ( motor.enabled ), tonumber ( motor.speed ), tonumber ( motor.maxTorque ) )
		end
	elseif joint.type == 'pulley' then
		local ground1 = xml.find ( joint, 'Ground1' )[1]
		local ground2 = xml.find ( joint, 'Ground2' )[1]

		local a1 = tob2Vec2 ( anchor1[1] )
		local a2 = tob2Vec2 ( anchor2[1] )
		local ga1 = tob2Vec2 ( ground1 )
		local ga2 = tob2Vec2 ( ground2 )

		j = self:CreateJointPulley ( body1, body2, collideConnected, a1, a2, ga1, ga2 )
	elseif joint.type == 'prismatic' then
		local a = tob2Vec2 ( anchor1[1] )
		local axis = tob2Vec2 ( xml.find ( joint, 'Axis' )[1] )

		j = self:CreateJointPrismatic ( body1, body2, collideConnected, a, axis )

		local limit = xml.find ( joint, 'Limit' )
		if limit then
			j:SetLimit ( toboolean ( limit.enabled ), tonumber ( limit.lowerTranslation ), tonumber ( limit.upperTranslation ) )
		end
		local motor = xml.find ( joint, 'Motor' )
		if motor then
			j:SetMotor ( toboolean ( motor.enabled ), tonumber ( motor.speed ), tonumber ( motor.maxForce ) )
		end
	else
		-- unknown joint type
		assert ( false, "Unkown joint type" )
	end

	-- could not generate joint
	assert ( j, "Could not generate joint" )

	return j
end

function World:CreateBodyXML ( body )
	local position = tob2Vec2 ( body.position )
	local angle = tonumber ( body.angle )
	local linearDamping = tonumber ( body.linearDamping )
	local angularDamping = tonumber ( body.angularDamping )
	local fixedRotation = toboolean ( body.fixedRotation )
	local isBullet = toboolean ( body.isBullet )
	local allowSleep = toboolean ( body.allowSleep )
	local isSleeping = toboolean ( body.isSleeping )

	local b = self:CreateBody ( body.name, position, angle, linearDamping, angularDamping, fixedRotation, isBullet, allowSleep, isSleeping )

	-- could not generate body
	assert ( b, "Could not generate body" )

	for i = 1, #body, 1 do
		local tag = xml.tag ( body[i] )

		if tag == 'Scale' then
			b.scale = body[i][1]
		elseif tag == "LinearVelocity" then
			local lv = tob2Vec2 ( body[i][1] )
			b.b2body:SetLinearVelocity ( lv )
		elseif tag == "AngularVelocity" then
			local av = body[i][1]
			b.b2body:SetAngularVelocity ( av )
		elseif tag == "Shape" then
			b:CreateShapeXML ( body[i] )
		elseif tag == "Goal" then
			b.goal = Goal:CreateXML ( body[i] )
			--b:SetGoal ( Goal:CreateXML ( body[i] ) )
		elseif tag == "Sprite" then
		    local s = body[i]
		    if s.alpha then
		        b.sprite.alpha = s.alpha
		    end
		    if s.depth then
		        b.sprite.depth = s.depth
		    end
		    if s.color then
		        b.sprite.color = toColor ( s.color )
		    end
		    if s.visible then
		        b.sprite.visible = toboolean ( s.visible )
		    end

            local image = xml.find ( s, 'Image' )
            if image then
                local x = tonumber ( image.imageX )
                local y = tonumber ( image.imageY )
                b:SetImage ( image[1], x, y )
            end
		else
			-- assert ( false, "Unkown tag in body" )
		end
	end

	-- body with no shapes
--	assert ( #b.shapes ~= 0, "Body with no shapes" )

	b:Paint ( )

	return b
end

function World:GetXML ( isCoroutine )
	local world = xml.new ( "World" )

	world.left = self.aabb.lowerBound.x
	world.top = self.aabb.lowerBound.y
	world.right = self.aabb.upperBound.x
	world.bottom = self.aabb.upperBound.y
	world.gravity = b2Vec2tostring ( self.gravity )
	world.doSleep = self.doSleep

	for i, v in pairs ( self.bodies ) do
		local body = v:GetXML ( )
		world:append ( body )

		if isCoroutine == true then
			coroutine.yield ( )
		end
	end
	for i, v in pairs ( self.joints ) do
		local joint = v:GetXML ( )
		world:append ( joint )

		if isCoroutine == true then
			coroutine.yield ( )
		end
	end

	return world
end


function World:SaveXML ( file, isCoroutine )
	local f = xml.new ( "XMLBox2D" )
	f.version = "1.0"
	f.created = os.date ( )

	local world = self:GetXML ( isCoroutine )

	f:append ( world )

	xml.save ( f, file )
end
